"""
Tools for working with dictionaries (and other Mappings).

If you are looking for more, check out the `lkj.iterables` module too 
(after all, dicts are iterables).

"""

from typing import Optional


def inclusive_subdict(d, include):
    """
    Returns a new dictionary with only the keys in `include`.

    Parameters:
    d (dict): The input dictionary.
    include (set): The set of keys to include in the new dictionary.

    Example:
    >>> inclusive_subdict({'a': 1, 'b': 2, 'c': 3}, {'a', 'c'})
    {'a': 1, 'c': 3}

    """
    return {k: d[k] for k in d.keys() & include}


def exclusive_subdict(d, exclude):
    """
    Returns a new dictionary with only the keys not in `exclude`.

    Parameters:
    d (dict): The input dictionary.
    exclude (set): The set of keys to exclude from the new dictionary.

    Example:
    >>> exclusive_subdict({'a': 1, 'b': 2, 'c': 3}, {'a', 'c'})
    {'b': 2}

    """
    return {k: d[k] for k in d.keys() - exclude}


# Note: There is a copy of truncate_dict_values in the ju package.
def truncate_dict_values(
    d: dict,
    *,
    max_list_size: Optional[int] = 2,
    max_string_size: Optional[int] = 66,
    middle_marker: str = "..."
) -> dict:
    """
    Returns a new dictionary with the same nested keys structure, where:
    - List values are reduced to a maximum size of max_list_size.
    - String values longer than max_string_size are truncated in the middle.

    Parameters:
    d (dict): The input dictionary.
    max_list_size (int, optional): Maximum size for lists. Defaults to 2.
    max_string_size (int, optional): Maximum length for strings. Defaults to None (no truncation).
    middle_marker (str, optional): String to insert in the middle of truncated strings. Defaults to '...'.

    Returns:
    dict: A new dictionary with truncated lists and strings.

    This can be useful when you have a large dictionary that you want to investigate,
    but printing/logging it takes too much space.

    Example:

    >>> large_dict = {'a': [1, 2, 3, 4, 5], 'b': {'c': [6, 7, 8, 9], 'd': 'A string like this that is too long'}, 'e': [10, 11]}
    >>> truncate_dict_values(large_dict, max_list_size=3, max_string_size=20)
    {'a': [1, 2, 3], 'b': {'c': [6, 7, 8], 'd': 'A string...too long'}, 'e': [10, 11]}

    You can use `None` to indicate "no max":

    >>> assert (
    ...     truncate_dict_values(large_dict, max_list_size=None, max_string_size=None)
    ...     == large_dict
    ... )

    """

    def truncate_string(value, max_len, marker):
        if max_len is None or len(value) <= max_len:
            return value
        half_len = (max_len - len(marker)) // 2
        return value[:half_len] + marker + value[-half_len:]

    kwargs = dict(
        max_list_size=max_list_size,
        max_string_size=max_string_size,
        middle_marker=middle_marker,
    )
    if isinstance(d, dict):
        return {k: truncate_dict_values(v, **kwargs) for k, v in d.items()}
    elif isinstance(d, list):
        return (
            [truncate_dict_values(v, **kwargs) for v in d[:max_list_size]]
            if max_list_size is not None
            else d
        )
    elif isinstance(d, str):
        return truncate_string(d, max_string_size, middle_marker)
    else:
        return d



"""Tools for importing"""


def import_object(dot_path: str):
    """Imports and returns an object from a dot string path.

    >>> f = import_object('os.path.join')
    >>> from os.path import join
    >>> f is join
    True

    """
    from importlib import import_module

    module_path, _, object_name = dot_path.rpartition(".")
    if not module_path:
        raise ImportError(f"{dot_path} does not contain a module path")
    module = import_module(module_path)
    try:
        return getattr(module, object_name)
    except AttributeError:
        raise ImportError(f"{object_name} is not found in {module_path}")


import sys
import importlib.util
import importlib.abc


class NamespaceForwardingLoader(importlib.abc.Loader):
    """
    A custom loader that forwards the import from a source namespace to a target namespace.

    Attributes:
        fullname (str): The full name of the module being imported.
        source_base (str): The base namespace to detect and forward from.
        target_base (str): The base namespace to forward the import to.

    Methods:
        load_module(fullname):
            Loads and returns the target module corresponding to the source module name.
    """

    def __init__(self, fullname, source_base, target_base):
        self.fullname = fullname
        self.source_base = source_base
        self.target_base = target_base

    def load_module(self, fullname):
        target_name = fullname.replace(self.source_base, self.target_base)

        # If the target module is already loaded, return it
        if target_name in sys.modules:
            return sys.modules[target_name]

        # Find the spec of the target module
        spec = importlib.util.find_spec(target_name)
        if spec is None:
            raise ImportError(f"No module named '{target_name}'")

        # Create and load the target module
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[fullname] = module
        return module


class NamespaceForwardingFinder(importlib.abc.MetaPathFinder):
    """
    A custom finder that detects when a module in the source namespace is being imported
    and forwards it to the target namespace.

    Attributes:
        source_base (str): The source namespace to detect and forward.
        target_base (str): The target namespace to forward the import to.

    Methods:
        find_spec(fullname, path, target=None):
            Finds and returns the spec for the target module corresponding to the source module name.
    """

    def __init__(self, source_base, target_base):
        self.source_base = source_base
        self.target_base = target_base

    def find_spec(self, fullname, path, target=None):
        if fullname.startswith(self.source_base):
            return importlib.util.spec_from_loader(
                fullname,
                NamespaceForwardingLoader(fullname, self.source_base, self.target_base),
            )
        return None


def register_namespace_forwarding(source_base, target_base):
    """
    Register the namespace forwarding from source_base to target_base.

    Args:
        source_base (str): The source namespace to forward.
        target_base (str): The target namespace to forward to.

    Usage:

        # if you put this code in the imbed.mdat package (say, containing a hcp module),
        >>> register_namespace_forwarding('imbed.mdat', 'imbed_data_prep')  # doctest: +SKIP

        # Then when you do

        >>> import imbed.mdat.hcp  # doctest: +SKIP

        You'll get the imbed_data_prep.hcp module.

    This function inserts the custom finder into sys.meta_path, enabling the
    dynamic import forwarding.
    """
    sys.meta_path.insert(0, NamespaceForwardingFinder(source_base, target_base))



"""Utils for logging."""

from typing import Callable, Tuple, Any, Optional, Union, Iterable
from functools import partial, wraps
from operator import attrgetter


# TODO: Verify and add test for line_prefix
# TODO: Merge with wrap_text_with_exact_spacing
# TODO: Add doctests for string
def wrapped_print(
    items: Union[str, Iterable],
    sep=", ",
    max_width=80,
    *,
    print_func=print,
    line_prefix: str = "",
):
    r"""
    Prints a string or list ensuring the total line width does not exceed `max_width`.

    If adding a new item would exceed this width, it starts a new line.

    Args:
        items (str or list): String or list of items to print.
        sep (str): The separator to use between items.
        max_width (int): The maximum width of each line. Default is 80.

    Example:

    >>> items = [
    ...     "item1", "item2", "item3", "item4", "item5", "item6", "item7", "item8",
    ...     "item9", "item10"
    ... ]
    >>> sep = ", "
    >>> wrapped_print(items, sep, max_width=30)
    item1, item2, item3, item4,
    item5, item6, item7, item8,
    item9, item10

    >>> items = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
    >>> sep = " - "
    >>> wrapped_print(items, sep, max_width=10)
    a - b - c
    - d - e -
    f - g - h
    - i - j

    Note that you have control over the `print_func`.
    This, for example, allows you to just return the string instead of printing it.

    >>> wrapped_print(items, sep, max_width=10, print_func=lambda x: x)
    'a - b - c\n- d - e -\nf - g - h\n- i - j'

    """

    if isinstance(items, str):
        return wrap_text_with_exact_spacing(
            items, max_width=max_width, print_func=print_func, line_prefix=line_prefix
        )
    else:
        import textwrap

        return print_func(
            line_prefix
            + textwrap.fill(
                sep.join(items), width=max_width, subsequent_indent=line_prefix
            )
        )


# TODO: Merge with wrapped_print
def wrap_text_with_exact_spacing(
    text, *, max_width=80, print_func=print, line_prefix: str = ""
):
    """
    Prints a string with word-wrapping to a maximum line length, while preserving all existing newlines
    exactly as they appear.

    Args:
    - text (str): The text to wrap and print.
    - max_width (int): The maximum width of each line (default is 88).
    """
    import textwrap

    # Split the text into lines, preserving the existing newlines
    lines = text.splitlines(keepends=True)

    # Wrap each line individually, preserving existing newlines and only adding new ones for wrapping
    wrapped_lines = [
        (
            textwrap.fill(
                line, width=max_width, replace_whitespace=False, drop_whitespace=False
            )
            if line.strip()
            else line_prefix + line
        )
        for line in lines
    ]

    # Join the wrapped lines back into a single text
    wrapped_text = "".join(wrapped_lines)

    return print_func(wrapped_text)


def print_with_timestamp(msg, *, refresh=None, display_time=True, print_func=print):
    """Prints with a timestamp and optional refresh.

    input: message, and possibly args (to be placed in the message string, sprintf-style

    output: Displays the time (HH:MM:SS), and the message

    use: To be able to track processes (and the time they take)

    """
    from datetime import datetime

    def hms_message(msg=""):
        t = datetime.now()
        return "({:02.0f}){:02.0f}:{:02.0f}:{:02.0f} - {}".format(
            t.day, t.hour, t.minute, t.second, msg
        )

    if display_time:
        msg = hms_message(msg)
    if refresh:
        print_func(msg, end="\r")
    else:
        print_func(msg)


print_progress = print_with_timestamp  # alias often used


def clog(condition, *args, log_func=print, **kwargs):
    """Conditional log

    >>> clog(False, "logging this")
    >>> clog(True, "logging this")
    logging this

    One common usage is when there's a verbose flag that allows the user to specify
    whether they want to log or not. Instead of having to litter your code with
    `if verbose:` statements you can just do this:

    >>> verbose = True  # say versbose is True
    >>> _clog = clog(verbose)  # makes a clog with a fixed condition
    >>> _clog("logging this")
    logging this

    You can also choose a different log function.
    Usually you'd want to use a logger object from the logging module,
    but for this example we'll just use `print` with some modification:

    >>> _clog = clog(verbose, log_func=lambda x: print(f"hello {x}"))
    >>> _clog("logging this")
    hello logging this

    """
    if not args and not kwargs:
        import functools

        return functools.partial(clog, condition, log_func=log_func)
    if condition:
        return log_func(*args, **kwargs)


def _calling_name(func_name: str, args: Tuple, kwargs: dict) -> str:
    return f"Calling {func_name}..."


def _done_calling_name(func_name: str, args: Tuple, kwargs: dict, result: Any) -> str:
    return f".... Done calling {func_name}"


def _always_log(func: Callable, args: Tuple, kwargs: dict) -> bool:
    """Return True no matter what"""
    return True


def log_calls(
    func: Callable = None,
    *,
    logger: Callable[[str], None] = print,
    ingress_msg: Callable[[str, Tuple, dict], str] = _calling_name,
    egress_msg: Callable[[str, Tuple, dict, Any], str] = _done_calling_name,
    func_name: Callable[[Callable], str] = attrgetter("__name__"),
    log_condition: Callable[[Callable, Tuple, dict], bool] = _always_log,
) -> Callable:
    """
    Decorator that adds logging before and after the function's call.

    Args:
        logger (callable): The logger function to use. Default is print.
        ingress_msg (callable): The message to log before calling the function.
            If it returns None, no message is logged.
            Default is "Calling {name}..." where name is the name of the function.
        egress_msg (callable): The message to log after the function call.
            If it returns None, no message is logged.
            Default is ".... Done".
        func_name (callable): The function to use to get the function's name.
        log_condition (callable): The condition for logging.
            Should be a callable that takes a function, args, and kwargs,
            and return a bool. If log_condition returns False, no logging is done.
            Default is _always_log.

    Tips:
        Use `logger=print_with_timestamp` to get timestamps in your logs.
        Use `lambda *args: None` as the ingress_msg or egress_msg to suppress logging.
        Use `logger=lambda x: None` to suppress all logging.

    Returns:
        The decorated function.

    Example:

    >>> @log_calls
    ... def add(a, b):
    ...     return a + b
    ...
    >>> add(2, 3)
    Calling add...
    .... Done calling add
    5

    >>> @log_calls(
    ...     logger=lambda x: print(f"LOG: {x}"),
    ...     ingress_msg=lambda name, *args: f"Start {name}!",
    ...     egress_msg=lambda *args: "End"
    ... )
    ... def multiply(a, b):
    ...     return a * b
    ...
    >>> multiply(2, 3)
    LOG: Start multiply!
    LOG: End
    6

    Sometimes, you want to dynamically control whether to log or not.
    This is what the `log_condition` parameter is for.
    One common use case is to log only if a flag is set in an instance.
    Since this is a common use case, we provide the `log_calls.instance_flag_is_set`
    helper function for this. You can use partial to set the flag attribute:

    >>> import functools
    >>> log_if_verbose_set_to_true = functools.partial(
    ...     log_calls.instance_flag_is_set, flag_attr='verbose'
    ... )

    Now if you have a class with a `verbose` attribute, you can use this helper function
    to log only if `verbose` is set:

    >>> class MyClass:
    ...     def __init__(self, verbose=False):
    ...         self.verbose = verbose
    ...
    ...     @log_calls(log_condition=log_if_verbose_set_to_true)
    ...     def foo(self):
    ...         print("Executing foo")
    ...
    >>> # Example usage
    >>> obj = MyClass(verbose=True)
    >>> obj.foo()  # This will log
    Calling foo...
    Executing foo
    .... Done calling foo

    But if verbose is set to `False`, no logging will be done:
    >>> obj = MyClass(verbose=False)
    >>> obj.foo()  # This will not log
    Executing foo

    """
    if func is None:
        return partial(
            log_calls,
            logger=logger,
            ingress_msg=ingress_msg,
            egress_msg=egress_msg,
            func_name=func_name,
            log_condition=log_condition,
        )

    @wraps(func)
    def wrapper(*args, **kwargs):
        if log_condition(func, args, kwargs):
            if ingress_msg:
                logger(ingress_msg(func_name(func), args, kwargs))
            result = func(*args, **kwargs)
            if egress_msg:
                logger(egress_msg(func_name(func), args, kwargs, result))
            return result
        else:
            return func(*args, **kwargs)

    return wrapper


def instance_flag_is_set(func, args, kwargs, flag_attr: str = "verbose"):
    """Check if the log flag is set to True in the instance."""
    # get the first argument if any, assuming it's the instance
    if flag_attr:
        instance = next(iter(args), None)
        if instance:  # assume it's the instance
            if getattr(instance, flag_attr, False):
                return True
    return False


log_calls.instance_flag_is_set = instance_flag_is_set

# if func is None:
#     return partial(
#         log_calls,
#         logger=logger,
#         ingress_msg=ingress_msg,
#         egress_msg=egress_msg,
#         func_name=func_name,
#     )
# else:

#     @wraps(func)
#     def wrapper(*args, **kwargs):
#         _func_name = func_name(func)
#         if (_in_msg := ingress_msg(_func_name, args, kwargs)) is not None:
#             logger(_in_msg)
#         result = func(*args, **kwargs)
#         if (_out_msg := egress_msg(_func_name, args, kwargs, result)) is not None:
#             logger(_out_msg)
#         return result

#     return wrapper


# --------------------------------------------------------------------------------------
# Error handling


from typing import Callable, Tuple, Any
from dataclasses import dataclass
import traceback
from typing import Callable, Any, Tuple
from functools import partial, wraps
from operator import attrgetter


@dataclass
class ErrorInfo:
    func: Callable
    error: Exception
    traceback: str
    locals: dict


def _dflt_msg_func(error_info: ErrorInfo) -> str:
    func_name = getattr(error_info.func, "__name__", "unknown")
    error_obj = error_info.error
    return f"Exiting from {func_name} with error: {error_obj}"


def dflt_error_info_processor(
    error_info: ErrorInfo,
    *,
    log_func=print,
    msg_func: Callable[[ErrorInfo], str] = _dflt_msg_func,
):
    return log_func(msg_func(error_info))


def return_error_info_on_error(
    func,
    *,
    caught_error_types: Tuple[Exception] = (Exception,),
    error_info_processor: Callable[[ErrorInfo], Any] = dflt_error_info_processor,
):
    """Decorator that returns traceback and local variables on error.

    This decorator is useful for debugging. It will catch any exceptions that occur
    in the decorated function, and return an ErrorInfo object with the traceback and
    local variables at the time of the error.

    :param func: The function to decorate.
    :param caught_error_types: The types of errors to catch.
    :param error_info_processor: A function that processes the ErrorInfo object.

    Tip: To parametrize this decorator, you can use a functools.partial function.

    Tip: You can have your error_info_processor persist the error info to a file or
    database, or send it to a logging service.

    >>> @return_error_info_on_error
    ... def foo(x, y=2):
    ...     return x / y
    ...
    >>> t = foo(1, 2)
    >>> assert t == 0.5
    >>> t = foo(1, y=0)
    Exiting from foo with error: division by zero
    >>> if isinstance(t, ErrorInfo):
    ...     assert isinstance(t.error, ZeroDivisionError)
    ...     hasattr(t, 'traceback')
    ...     assert t.locals['args'] == (1,)
    ...     assert t.locals['kwargs'] == {'y': 0}

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except caught_error_types as error_obj:
            error_info = ErrorInfo(func, error_obj, traceback.format_exc(), locals())
            return error_info_processor(error_info)

    return wrapper



"""
Lightweight Kit Jumpstart
"""

from lkj.iterables import (
    compare_sets,  # Compare two iterables and return common, left-only, and right-only elements
    index_of,  # Get the index of an element in an iterable
    get_by_value,  # Get a dictionary from a list of dictionaries by a field value
)
from lkj.funcs import mk_factory
from lkj.dicts import truncate_dict_values, inclusive_subdict, exclusive_subdict
from lkj.filesys import get_app_data_dir, get_watermarked_dir, enable_sourcing_from_file
from lkj.strings import (
    indent_lines,  # Indent all lines of a string
    most_common_indent,  # Get the most common indent of a multiline string
    regex_based_substitution,
    truncate_string_with_marker,  # Truncate a string to a maximum length, inserting a marker in the middle.
    unique_affixes,  # Get unique prefixes or suffixes of a list of strings
    camel_to_snake,  # Convert CamelCase to snake_case
    snake_to_camel,  # Convert snake_case to CamelCase
    fields_of_string_format,  # Extract field names from a string format
    fields_of_string_formats,  # Extract field names from an iterable of string formats
)
from lkj.loggers import (
    print_with_timestamp,
    print_progress,
    log_calls,
    clog,
    ErrorInfo,
    return_error_info_on_error,
    wrapped_print,
)
from lkj.importing import import_object, register_namespace_forwarding
from lkj.chunking import chunk_iterable, chunker
from lkj.misc import identity, value_in_interval

ddir = lambda obj: list(filter(lambda x: not x.startswith("_"), dir(obj)))


def user_machine_id():
    """Get an ID for the current computer/user that calls this function."""
    return __import__("platform").node()


def add_attr(attr_name: str, attr_val: str = None, obj=None):
    """Add an attribute to an object.

    If no object is provided, return a partial function that takes an object as its
    argument.
    If no attribute value is provided, return a partial function that takes an
    attribute value as its argument.
    If no object or attribute value is provided, return a partial function that takes
    both an object and an attribute value as its arguments.
    If all arguments are provided, add the attribute to the object and return the
    object.

    :param attr_name: The name of the attribute to add.
    :param attr_val: The value of the attribute to add.
    :param obj: The object to which to add the attribute.
    :return: The object with the attribute added, or a partial function that takes an
    object and/or an attribute value as its argument(s).

    >>> def generic_func(*args, **kwargs):
    ...     return args, kwargs
    ...
    >>> generic_func.__name__
    'generic_func'
    >>>
    >>> _ = add_attr('__name__', 'my_func', generic_func);
    >>> generic_func.__name__
    'my_func'
    >>>
    >>>
    >>> add_name = add_attr('__name__')
    >>> add_doc = add_attr('__doc__')
    >>>
    >>> @add_name('my_func')
    ... @add_doc('This is my function.')
    ... def f(*args, **kwargs):
    ...     return args, kwargs
    ...
    >>> f.__name__
    'my_func'
    >>> f.__doc__
    'This is my function.'

    """
    if obj is None:
        from functools import partial

        if attr_val is None:
            return partial(add_attr, attr_name)
        return partial(add_attr, attr_name, attr_val)
    if attr_val is not None:
        setattr(obj, attr_name, attr_val)
    return obj


def add_as_attribute_of(obj, name=None):
    """Decorator that adds a function as an attribute of a container object ``obj``.

    If no ``name`` is given, the ``__name__`` of the function will be used, with a
    leading underscore removed. This is useful for adding helper functions to main
    "container" functions without polluting the namespace of the module, at least
    from the point of view of imports and tab completion.

    >>> def foo():
    ...    pass
    >>>
    >>> @add_as_attribute_of(foo)
    ... def helper():
    ...    pass
    >>> hasattr(foo, 'helper')
    True
    >>> callable(foo.helper)
    True

    In reality, any object that has a ``__name__`` can be added to the attribute of
    ``obj``, but the intention is to add helper functions to main "container" functions.

    Note that if the name of the function starts with an underscore, it will be removed
    before adding it as an attribute of ``obj``.

    >>> @add_as_attribute_of(foo)
    ... def _helper():
    ...    pass
    >>> hasattr(foo, 'helper')
    True

    This is useful for adding helper functions to main "container" functions without
    polluting the namespace of the module, at least from the point of view of imports
    and tab completion. But if you really want to add a function with a leading
    underscore, you can do so by specifying the name explicitly:

    >>> @add_as_attribute_of(foo, name='_helper')
    ... def _helper():
    ...    pass
    >>> hasattr(foo, '_helper')
    True

    Of course, you can give any name you want to the attribute:

    >>> @add_as_attribute_of(foo, name='bar')
    ... def _helper():
    ...    pass
    >>> hasattr(foo, 'bar')
    True

    :param obj: The object to which the function will be added as an attribute
    :param name: The name of the attribute to add the function to. If not given, the

    """

    def _decorator(f):
        attrname = name or f.__name__
        if not name and attrname.startswith("_"):
            attrname = attrname[1:]  # remove leading underscore
        setattr(obj, attrname, f)
        return f

    return _decorator


def get_caller_package_name(default=None):
    """Return package name of caller

    See: https://github.com/i2mint/i2mint/issues/1#issuecomment-1479416085
    """
    import inspect

    try:
        stack = inspect.stack()
        caller_frame = stack[1][0]
        return inspect.getmodule(caller_frame).__name__.split(".")[0]
    except Exception as error:
        return default



"""Tools for chunking (segumentation, batching, slicing, etc.)"""

from itertools import zip_longest, chain, islice

from typing import (
    Iterable,
    Union,
    Dict,
    List,
    Tuple,
    Mapping,
    TypeVar,
    Iterator,
    Callable,
    Optional,
    T,
)

KT = TypeVar("KT")  # there's a typing.KT, but pylance won't allow me to use it!
VT = TypeVar("VT")  # there's a typing.VT, but pylance won't allow me to use it!


def chunk_iterable(
    iterable: Union[Iterable[T], Mapping[KT, VT]],
    chk_size: int,
    *,
    chunk_type: Optional[Callable[..., Union[Iterable[T], Mapping[KT, VT]]]] = None,
) -> Iterator[Union[List[T], Tuple[T, ...], Dict[KT, VT]]]:
    """
    Divide an iterable into chunks/batches of a specific size.

    Handles both mappings (e.g. dicts) and non-mappings (lists, tuples, sets...)
    as you probably expect it to (if you give a dict input, it will chunk on the
    (key, value) items and return dicts of these).
    Thought note that you always can control the type of the chunks with the
    `chunk_type` argument.

    Args:
        iterable: The iterable or mapping to divide.
        chk_size: The size of each chunk.
        chunk_type: The type of the chunks (list, tuple, set, dict...).

    Returns:
        An iterator of dicts if the input is a Mapping, otherwise an iterator
        of collections (list, tuple, set...).

    Examples:
        >>> list(chunk_iterable([1, 2, 3, 4, 5], 2))
        [[1, 2], [3, 4], [5]]

        >>> list(chunk_iterable((1, 2, 3, 4, 5), 3, chunk_type=tuple))
        [(1, 2, 3), (4, 5)]

        >>> list(chunk_iterable({"a": 1, "b": 2, "c": 3}, 2))
        [{'a': 1, 'b': 2}, {'c': 3}]

        >>> list(chunk_iterable({"x": 1, "y": 2, "z": 3}, 1, chunk_type=dict))
        [{'x': 1}, {'y': 2}, {'z': 3}]
    """
    if isinstance(iterable, Mapping):
        if chunk_type is None:
            chunk_type = dict
        it = iter(iterable.items())
        for first in it:
            yield {
                key: value for key, value in chain([first], islice(it, chk_size - 1))
            }
    else:
        if chunk_type is None:
            if isinstance(iterable, (list, tuple, set)):
                chunk_type = type(iterable)
            else:
                chunk_type = list
        it = iter(iterable)
        for first in it:
            yield chunk_type(chain([first], islice(it, chk_size - 1)))


def chunker(
    a: Iterable[T], chk_size: int, *, include_tail: bool = True
) -> Iterator[Tuple[T, ...]]:
    """
    Chunks an iterable into non-overlapping chunks of size `chk_size`.

    Note: This chunker is simpler, but also less efficient than `chunk_iterable`.
    It does have the extra `include_tail` argument, though.
    Though note that you can get the effect of `include_tail=False` in `chunk_iterable`
    by using `filter(lambda x: len(x) == chk_size, chunk_iterable(...))`.

    Args:
        a: The iterable to be chunked.
        chk_size: The size of each chunk.
        include_tail: If True, includes the remaining elements as the last chunk
                      even if they are fewer than `chk_size`. Defaults to True.

    Returns:
        An iterator of tuples, where each tuple is a chunk of size `chk_size`
        (or fewer elements if `include_tail` is True).

    Examples:
        >>> list(chunker(range(8), 3))
        [(0, 1, 2), (3, 4, 5), (6, 7)]
        >>> list(chunker(range(8), 3, include_tail=False))
        [(0, 1, 2), (3, 4, 5)]
    """
    it = iter(a)
    if include_tail:
        sentinel = object()
        for chunk in zip_longest(*([it] * chk_size), fillvalue=sentinel):
            yield tuple(item for item in chunk if item is not sentinel)
    else:
        yield from zip(*([it] * chk_size))



"""Miscellaneous tools."""

from operator import attrgetter, ge, gt, le, lt
from functools import partial
from typing import Callable, T, Optional, Any


def identity(x):
    return x


def value_in_interval(
    x: Any = None,
    /,
    *,
    get_val: Callable[[Any], T] = identity,
    min_val: Optional[T] = None,
    max_val: Optional[T] = None,
    is_minimum: Callable[[T, T], bool] = ge,
    is_maximum: Callable[[T, T], bool] = lt,
):
    """

    >>> from operator import itemgetter, le
    >>> f = value_in_interval(get_val=itemgetter('date'), min_val=2, max_val=8)
    >>> d = [{'date': 1}, {'date': 2}, {'date': 3, 'x': 7}, {'date': 8}, {'date': 9}]
    >>> list(map(f, d))
    [False, True, True, False, False]

    The default `is_maximum` is `lt` (i.e. lambda x: x < max_val). If you want to
    use `le` (i.e. lambda x: x <= max_val) you can change the is_maximum argument:

    >>> ff = value_in_interval(
    ...     get_val=itemgetter('date'), min_val=2, max_val=8, is_maximum=le
    ... )
    >>> list(map(ff, d))
    [False, True, True, True, False]
    """
    if x is None:
        kwargs = locals()
        kwargs.pop("x")
        return partial(value_in_interval, **kwargs)
    else:
        val = get_val(x)
        if min_val is not None and not is_minimum(val, min_val):
            return False
        if max_val is not None and not is_maximum(val, max_val):
            return False
        return True



"""Utils for strings"""

import re


def indent_lines(string: str, indent: str, *, line_sep="\n") -> str:
    r"""
    Indent each line of a string.

    :param string: The string to indent.
    :param indent: The string to use for indentation.
    :return: The indented string.

    >>> print(indent_lines('This is a test.\nAnother line.', ' ' * 8))
            This is a test.
            Another line.
    """
    return line_sep.join(indent + line for line in string.split(line_sep))


def most_common_indent(string: str, ignore_first_line=False) -> str:
    r"""
    Find the most common indentation in a string.

    :param string: The string to analyze.
    :param ignore_first_line: Whether to ignore the first line when determining the
        indentation. Default is False. One case where you want True is when using python
        triple quotes (as in docstrings, for example), since the first line often has
        no indentation (from the point of view of the string, in this case.
    :return: The most common indentation string.

    Examples:

    >>> most_common_indent('    This is a test.\n    Another line.')
    '    '
    """
    indents = re.findall(r"^( *)\S", string, re.MULTILINE)
    n_lines = len(indents)
    if ignore_first_line and n_lines > 1:
        # if there's more than one line, ignore the indent of the first
        indents = indents[1:]
    return max(indents, key=indents.count)


from string import Formatter

formatter = Formatter()


def fields_of_string_format(template):
    return [
        field_name for _, field_name, _, _ in formatter.parse(template) if field_name
    ]


def fields_of_string_formats(templates, *, aggregator=set):
    """
    Extract all unique field names from the templates in _github_url_templates using string.Formatter.

    Args:
        templates (list): A list of dictionaries containing 'template' keys.

    Returns:
        list: A sorted list of unique field names found in the templates.

    Example:
        >>> templates = ['{this}/and/{that}', 'and/{that}/is/an/{other}']
        >>> sorted(fields_of_string_formats(templates))
        ['other', 'that', 'this']
    """

    def field_names():
        for template in templates:
            yield from fields_of_string_format(template)

    return aggregator(field_names())


import re

# Compiled regex to handle camel case to snake case conversions, including acronyms
_camel_to_snake_re = re.compile(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")


def camel_to_snake(camel_string):
    """
    Convert a CamelCase string to snake_case. Useful for converting class
    names to variable names.

    Args:
        camel_string (str): The CamelCase string to convert.

    Returns:
        str: The converted snake_case string.

    Examples:
        >>> camel_to_snake('BasicParseTest')
        'basic_parse_test'
        >>> camel_to_snake('HTMLParser')
        'html_parser'
        >>> camel_to_snake('CamelCaseExample')
        'camel_case_example'

        Note that acronyms are handled correctly:

        >>> camel_to_snake('XMLHttpRequestTest')
        'xml_http_request_test'
    """
    return _camel_to_snake_re.sub(r"_\1", camel_string).lower()


def snake_to_camel(snake_string):
    """
    Convert a snake_case string to CamelCase. Useful for converting variable
    names to class names.

    Args:
        snake_string (str): The snake_case string to convert.

    Returns:
        str: The converted CamelCase string.

    Examples:

        >>> snake_to_camel('complex_tokenizer')
        'ComplexTokenizer'
        >>> snake_to_camel('simple_example_test')
        'SimpleExampleTest'

        Note that acronyms are capitalized correctly:

        >>> snake_to_camel('xml_http_request_test')
        'XmlHttpRequestTest'
    """
    return "".join(word.capitalize() or "_" for word in snake_string.split("_"))


# Note: Vendored in i2.multi_objects and dol.util
def truncate_string_with_marker(
    s, *, left_limit=15, right_limit=15, middle_marker="..."
):
    """
    Truncate a string to a maximum length, inserting a marker in the middle.

    If the string is longer than the sum of the left_limit and right_limit,
    the string is truncated and the middle_marker is inserted in the middle.

    If the string is shorter than the sum of the left_limit and right_limit,
    the string is returned as is.

    >>> truncate_string_with_marker('1234567890')
    '1234567890'

    But if the string is longer than the sum of the limits, it is truncated:

    >>> truncate_string_with_marker('1234567890', left_limit=3, right_limit=3)
    '123...890'
    >>> truncate_string_with_marker('1234567890', left_limit=3, right_limit=0)
    '123...'
    >>> truncate_string_with_marker('1234567890', left_limit=0, right_limit=3)
    '...890'

    If you're using a specific parametrization of the function often, you can
    create a partial function with the desired parameters:

    >>> from functools import partial
    >>> truncate_string = partial(truncate_string_with_marker, left_limit=2, right_limit=2, middle_marker='---')
    >>> truncate_string('1234567890')
    '12---90'
    >>> truncate_string('supercalifragilisticexpialidocious')
    'su---us'

    """
    if len(s) <= left_limit + right_limit:
        return s
    elif right_limit == 0:
        return s[:left_limit] + middle_marker
    elif left_limit == 0:
        return middle_marker + s[-right_limit:]
    else:
        return s[:left_limit] + middle_marker + s[-right_limit:]


# TODO: Generalize so that it can be used with regex keys (not escaped)
def regex_based_substitution(replacements: dict, regex=None, s: str = None):
    """
    Construct a substitution function based on an iterable of replacement pairs.

    :param replacements: An iterable of (replace_this, with_that) pairs.
    :type replacements: iterable[tuple[str, str]]
    :return: A function that, when called with a string, will perform all substitutions.
    :rtype: Callable[[str], str]

    The function is meant to be used with ``replacements`` as its single input,
    returning a ``substitute`` function that will carry out the substitutions
    on an input string.

    >>> replacements = {'apple': 'orange', 'banana': 'grape'}
    >>> substitute = regex_based_substitution(replacements)
    >>> substitute("I like apple and bananas.")
    'I like orange and grapes.'

    You have access to the ``replacements`` and ``regex`` attributes of the
    ``substitute`` function:

    >>> substitute.replacements
    {'apple': 'orange', 'banana': 'grape'}

    """
    import re
    from functools import partial

    if regex is None and s is None:
        replacements = dict(replacements)

        if not replacements:  # if replacements iterable is empty.
            return lambda s: s  # return identity function

        regex = re.compile("|".join(re.escape(key) for key in replacements.keys()))

        substitute = partial(regex_based_substitution, replacements, regex)
        substitute.replacements = replacements
        substitute.regex = regex
        return substitute
    else:
        return regex.sub(lambda m: replacements[m.group(0)], s)


from typing import Callable, Iterable, Sequence


class TrieNode:
    def __init__(self):
        self.children = {}
        self.count = 0  # Number of times this node is visited during insertion
        self.is_end = False  # Indicates whether this node represents the end of an item


def identity(x):
    return x


def unique_affixes(
    items: Iterable[Sequence],
    suffix: bool = False,
    *,
    egress: Callable = None,
    ingress: Callable = identity,
) -> Iterable[Sequence]:
    """
    Returns a list of unique prefixes (or suffixes) for the given iterable of sequences.
    Raises a ValueError if duplicates are found.

    Parameters:
    - items: Iterable of sequences (e.g., list of strings).
    - suffix: If True, finds unique suffixes instead of prefixes.
    - ingress: Callable to preprocess each item. Default is identity function.
    - egress: Callable to postprocess each affix. Default is appropriate function based on item type.
      Usually, ingress and egress are inverses of each other.

    >>> unique_affixes(['apple', 'ape', 'apricot', 'banana', 'band', 'bandana'])
    ['app', 'ape', 'apr', 'bana', 'band', 'banda']

    >>> unique_affixes(['test', 'testing', 'tester'])
    ['test', 'testi', 'teste']

    >>> unique_affixes(['test', 'test'])
    Traceback (most recent call last):
    ...
    ValueError: Duplicate item detected: test

    >>> unique_affixes(['abc', 'abcd', 'abcde'])
    ['abc', 'abcd', 'abcde']

    >>> unique_affixes(['a', 'b', 'c'])
    ['a', 'b', 'c']

    >>> unique_affixes(['x', 'xy', 'xyz'])
    ['x', 'xy', 'xyz']

    >>> unique_affixes(['can', 'candy', 'candle'])
    ['can', 'candy', 'candl']

    >>> unique_affixes(['flow', 'flower', 'flight'])
    ['flow', 'flowe', 'fli']

    >>> unique_affixes(['ation', 'termination', 'examination'], suffix=True)
    ['ation', 'rmination', 'amination']

    >>> import functools
    >>> ingress = functools.partial(str.split, sep='.')
    >>> egress = '.'.join
    >>> items = ['here.and.there', 'here.or.there', 'here']
    >>> unique_affixes(items, ingress=ingress, egress=egress)
    ['here.and', 'here.or', 'here']

    """
    items = list(map(ingress, items))

    # Determine the default egress function based on item type
    if egress is None:
        if all(isinstance(item, str) for item in items):
            # Items are strings; affixes are lists of characters
            def egress(affix):
                return "".join(affix)

        else:
            # Items are sequences (e.g., lists); affixes are lists
            def egress(affix):
                return affix

    # If suffix is True, reverse the items
    if suffix:
        items = [item[::-1] for item in items]

    # Build the trie and detect duplicates
    root = TrieNode()
    for item in items:
        node = root
        for element in item:
            if element not in node.children:
                node.children[element] = TrieNode()
            node = node.children[element]
            node.count += 1
        # At the end of the item
        if node.is_end:
            # Duplicate detected
            if suffix:
                original_item = item[::-1]
            else:
                original_item = item
            original_item = egress(original_item)
            raise ValueError(f"Duplicate item detected: {original_item}")
        node.is_end = True

    # Find the minimal unique prefixes/suffixes
    affixes = []
    for item in items:
        node = root
        affix = []
        for element in item:
            node = node.children[element]
            affix.append(element)
            if node.count == 1:
                break
        if suffix:
            affix = affix[::-1]
        affixes.append(affix)

    # Postprocess affixes using egress
    affixes = list(map(egress, affixes))
    return affixes



"""Tools for working with functions."""

from functools import partial

mk_factory = partial(partial, partial)
mk_factory.__doc__ = """
Create a factory for functions by partially applying the first argument.

This higher-order function allows you to fix the first argument of a given function, creating a new
function that requires fewer arguments. It's particularly useful for creating customized functions
on the fly without using lambdas.

Example:

>>> from operator import mul
>>> double = mk_factory(mul)(2)
>>> double(5)
10
"""

# --------------------------------------------------------------------------------------
# operator: Fixing an operand
#
# You can use our trick to fix the (first) operand of an operation. This is useful,
# when, for example, you want to avoid using a `lambda` function in some situation
# (say, because [lambda functions are not pickle-serializable](https://realpython.com/python-pickle-module/)).

import operator

greater_than = mk_factory(operator.lt)
greater_than.__doc__ = """
Create a factory for a function that returns True if the argument is greater than the fixed value.

So, for instance, instead of using a lambda x: 10 < x you can do:

>>> greater_than_10 = greater_than(10)
>>> greater_than_10(11)
True
>>> greater_than_10(9)
False

This is even more useful when you need to define a bunch of greater_than functions, or do so dynamically!
"""

# Other examples:
less_than = mk_factory(operator.gt)
less_than.__doc__ = """
Create a factory for a function that returns True if the argument is less than the fixed value.

Example:

>>> less_than_10 = less_than(10)
>>> less_than_10(9)
True
>>> less_than_10(11)
False
"""

greater_or_equal_to = mk_factory(operator.le)
greater_or_equal_to.__doc__ = """
Create a factory for a function that returns True if the argument is greater than or equal to the fixed value.

Example:

>>> ge_10 = greater_or_equal_to(10)
>>> ge_10(10)
True
>>> ge_10(9)
False
"""

less_or_equal_to = mk_factory(operator.ge)
less_or_equal_to.__doc__ = """
Create a factory for a function that returns True if the argument is less than or equal to the fixed value.

Example:

>>> le_10 = less_or_equal_to(10)
>>> le_10(10)
True
>>> le_10(11)
False
"""

# --------------------------------------------------------------------------------------
# os.path.join: A Path Generator for Files
#
# Imagine working with a large project where you frequently need to generate file paths.
# With mk_factory, creating a path generator becomes effortless.

import os

mk_abs_path_maker = mk_factory(os.path.join)
mk_abs_path_maker.__doc__ = """
Create a factory for functions that join paths with a fixed root.

Example:

>>> abs_path = mk_abs_path_maker('/my/root')
>>> abs_path('a', 'b')
'/my/root/a/b'

Now abs_path behaves like os.path.join but with the root path '/my/root' fixed.
"""

# --------------------------------------------------------------------------------------
# map: Transforming Functions for Iterables: The "Iterizer"
#
# The "iterizer" pattern allows us to take any function and apply it across each element of an iterable.

iterizer = mk_factory(map)
iterizer.__doc__ = """
Create a factory that returns a function applying a given function to each element of an iterable.

Example:

>>> times_two = iterizer(lambda x: x * 2)
>>> list(times_two([1, 2, 3]))
[2, 4, 6]

This is useful when processing collections of data where each item must undergo the same transformation.
"""

# --------------------------------------------------------------------------------------
# filter: Filtering Collections: The "Filterizer"
#
# Create a factory to transform any function into a filter for iterables.

filterizer = mk_factory(filter)
filterizer.__doc__ = """
Create a factory that returns a function filtering elements of an iterable based on a predicate function.

Example:

>>> only_evens = filterizer(lambda x: x % 2 == 0)
>>> list(only_evens(range(10)))
[0, 2, 4, 6, 8]

Such "filterized" functions are powerful for handling datasets where specific criteria need to be applied.
"""

# --------------------------------------------------------------------------------------
# zip: Custom Zippers for Pairing Data
#
# With zipper, we can create functions that zip multiple iterables with predefined keys or labels.

zipper = mk_factory(zip)
zipper.__doc__ = """
Create a factory that returns a function zipping iterables with a fixed first iterable.

Example:

>>> my_zipper = zipper(['the', 'first', 'three'])
>>> dict(my_zipper([1, 2]))
{'the': 1, 'first': 2}
>>> dict(my_zipper([10, 20, 30]))
{'the': 10, 'first': 20, 'three': 30}

This approach is particularly useful for converting data into dictionaries with meaningful keys for easier access and readability.
"""

# --------------------------------------------------------------------------------------
# zip: Zipping with an Infinite Iterator for Unique Key Generation
#
# Use zipper with an infinite iterator to provide a continuous stream of unique keys to pair with values dynamically.

from itertools import count


def zip_with_filenames(start_idx=1):
    """
    Create a function that zips an iterable with file names like 'file_001', 'file_002', etc.

    Example:

    >>> zip_files = zip_with_filenames()
    >>> dict(zip_files([1, 2]))
    {'file_001': 1, 'file_002': 2}
    >>> dict(zip_files(['never', 'say', 'never']))
    {'file_004': 'never', 'file_005': 'say', 'file_006': 'never'}

    Using count ensures that each label is unique and ordered sequentially.
    """
    filenames = map("file_{:03.0f}".format, count(start_idx))
    return zipper(filenames)


# --------------------------------------------------------------------------------------
# regex: Regex Matchers and Extractors
#
# For tasks involving regular expressions, mk_factory can simplify creating matchers and extractors.

import re

matcher = mk_factory(re.match)
matcher.__doc__ = """
Create a factory that returns a function to match a pattern at the beginning of a string.

Example:

>>> is_email = matcher(r'^[\\w\\.-]+@[\\w\\.-]+\\.\\w+$')
>>> bool(is_email('example@example.com'))
True
>>> is_email('not an email') is None
True

With is_email, you have a quick function for checking email format.
"""

extractor = mk_factory(re.findall)
extractor.__doc__ = """
Create a factory that returns a function to find all non-overlapping matches of a pattern in a string.

Example:

>>> extract_numbers = extractor(r'[\\d\\.]+')
>>> extract_numbers('7 times 6 is 42.0')
['7', '6', '42.0']

These functions streamline text processing workflows, making pattern matching and extraction highly reusable and easy to set up.
"""

# --------------------------------------------------------------------------------------
# reduce: Reduction Functions for Aggregation
#
# Reduction functions allow aggregation of elements, useful for calculating products, sums, or even concatenating strings.

from functools import reduce
from operator import mul, add

reducer = mk_factory(reduce)
reducer.__doc__ = """
Create a factory that returns a function to reduce an iterable using a given binary function.

Example:

>>> product = reducer(mul)
>>> product([2, 3, 4])
24

>>> concatinator = reducer(add)
>>> concatinator(['a', 'b', 'c'])
'abc'

This approach lets you create highly readable, purpose-specific aggregation functions with minimal code.
"""

# --------------------------------------------------------------------------------------
# Creating a Function Composer for Pipelines
#
# Build a function composer that lets you apply a sequence of transformations or operations on data.

composer = reducer(lambda f, g: lambda *args, **kwargs: f(g(*args, **kwargs)))
composer.__doc__ = """
Create a function composer that composes a sequence of functions into a single function.

Example:

>>> exclaimed_title = composer([lambda x: x + '!', str.title])
>>> exclaimed_title('hello world')
'Hello World!'

This is especially useful in data pipelines, where data needs to pass through a series of transformations.
"""

# --------------------------------------------------------------------------------------
# Additional Examples


# Example using map with os.path.join to create full file paths
def full_path_maker(base_directory):
    """
    Create a function that generates full file paths given filenames.

    Example:

    >>> full_path = full_path_maker('/base/directory')
    >>> list(full_path(['file1', 'file2']))
    ['/base/directory/file1', '/base/directory/file2']
    """
    abs_path = mk_abs_path_maker(base_directory)
    return iterizer(abs_path)


# Example of using greater_than in a filter
def filter_greater_than(threshold, data):
    """
    Filter elements in data that are greater than the threshold.

    Example:

    >>> filter_greater_than_5 = filter_greater_than(5, range(10))
    >>> list(filter_greater_than_5)
    [6, 7, 8, 9]
    """
    gt = greater_than(threshold)
    return filter(gt, data)
